import contextlib
from typing import Annotated, AsyncIterator

import asyncpg
from fastapi import Depends, Request


@contextlib.asynccontextmanager
async def create_pool(dsn: str) -> AsyncIterator[asyncpg.Pool]:
    """
    Create a postgres connection pool.

    Arguments:
    - dsn: Connection arguments specified using as a single string in
           the following format:
           `postgres://user:pass@host:port/database?option=value`.
    """

    async def init_pool(conn: asyncpg.Connection):
        # Need the extension before we register, so do this here.
        await conn.execute("CREATE EXTENSION IF NOT EXISTS vector")

        from pgvector.asyncpg import register_vector

        await register_vector(conn)

    pool = await asyncpg.create_pool(dsn, init=init_pool)
    yield pool
    await pool.close()


def _pg_pool(request: Request) -> asyncpg.Pool:
    return request.state.pg_pool


PgPoolDep = Annotated[asyncpg.Pool, Depends(_pg_pool)]


async def _pg_connection(pool: PgPoolDep) -> asyncpg.Connection:
    async with pool.acquire() as connection:
        yield connection


PgConnectionDep = Annotated[asyncpg.Connection, Depends(_pg_connection)]



from typing import Any, Optional

from fastapi.routing import APIRoute
from pydantic import PostgresDsn, RedisDsn, ValidationInfo, field_validator
from pydantic_core import Url
from pydantic_settings import BaseSettings, SettingsConfigDict

from dewy.constants import Environment

# See https://github.com/zhanymkanov/fastapi-best-practices#10-use-pydantics-basesettings-for-configs


class Config(BaseSettings):
    """Application configuration, parsed from environment variables."""

    model_config = SettingsConfigDict(
        env_file=".env",
        env_file_encoding="utf-8",
    )

    SERVE_ADMIN_UI: bool = True
    """If true, serve the admin UI."""

    DB: PostgresDsn
    """The Postgres database to connect to."""

    APPLY_MIGRATIONS: bool = True
    """Whether migrations should be applied to the database.

    This should be set to false if there are multiple services running to prevent
    concurrent migrations.
    """

    ENVIRONMENT: Environment = Environment.PRODUCTION
    """The environment the application is running in."""

    REDIS: Optional[RedisDsn] = None
    """The Redis service to use for queueing, indexing and document storage."""

    EMBEDDING_MODEL: str = ""
    """The embedding model to use.

    This is a string of the form `<kind>[:<model>]` with the following options:

    1. `local[:<path or repository>]` -- Run a model locally. If this is a path
       it will attempt to load a model from that location. Otherwise, it should
       be a Hugging Face repository from which to retrieve the model.
    2. `openai[:<name>]` -- The named OpenAI model. `OPENAI_API_KEY` must be set.
    3. `ollama:<name>` -- The named Ollama model. `OLLAMA_BASE_URL` must be set.

    In each of these cases, you can omit the second part for the default model of the
    given kind.

    If unset, this will default to `"openai"` if an OpenAI API KEY is available and
    otherwise will use `"local"`.

    NOTE: Changing embedding models is not currently supported.
    """

    LLM_MODEL: str = ""
    """The LLM model to use.

    This is a string of the form `<kind>:<model>` with the following options:

    1. `local[:<path or repository>]` -- Run a model locally. If this is a path
       it will attempt to load a model from that location. Otherwise, it should
       be a Hugging Face repository from which to retrieve the model.
    2. `openai[:<name>]` -- The named OpenAI model. `OPENAI_API_KEY` must be set.
    3. `ollama:<name>` -- The named Ollama model. `OLLAMA_BASE_URL` must be set.

    In each of these cases, you can omit the second part for the default model of the
    given kind.

    If unset, this will default to `"openai"` if an OpenAI API KEY is available and
    otherwise will use `"local"`.
    """

    OPENAI_API_KEY: Optional[str] = None
    """ The OpenAI API Key to use for OpenAI models.

    This is required for using openai models.
    """

    OLLAMA_BASE_URL: Optional[Url] = None
    """The Base URL for Ollama.

    This is required for using ollama models.
    """

    @field_validator("OLLAMA_BASE_URL")
    def validate_ollama_base_url(cls, v, info: ValidationInfo):
        MODELS = ["LLM_MODEL", "EMBEDDING_MODEL"]
        if v is None:
            for model in MODELS:
                context = info.context
                if context:
                    value = context.get(model, "")
                    if value.startswith("ollama"):
                        raise ValueError(
                            f"{info.field_name} must be set to use '{model}={value}'"
                        )
        return v


settings = Config()


def convert_snake_case_to_camel_case(string: str) -> str:
    """Convert snake case to camel case"""

    words = string.split("_")
    return words[0] + "".join(word.title() for word in words[1:])


def custom_generate_unique_id_function(route: APIRoute) -> str:
    """Custom function to generate unique id for each endpoint"""

    return convert_snake_case_to_camel_case(route.name)


API_DESCRIPTION: str = """This API allows ingesting and retrieving knowledge.

Knowledge comes in a variety of forms -- text, image, tables, etc. and
from a variety of sources -- documents, web pages, audio, etc.
"""

app_configs: dict[str, Any] = {
    "title": "Dewy Knowledge Base API",
    "summary": "Knowledge curation for Retrieval Augmented Generation",
    "description": API_DESCRIPTION,
    "servers": [
        {"url": "http://localhost:8000", "description": "Local server"},
    ],
    "generate_unique_id_function": custom_generate_unique_id_function,
}

if not settings.ENVIRONMENT.is_debug:
    app_configs["openapi_url"] = None  # hide docs



from typing import Annotated, List

from fastapi import APIRouter, Path
from loguru import logger

from dewy.collection.models import Collection, CollectionCreate
from dewy.common.collection_embeddings import get_dimensions
from dewy.common.db import PgConnectionDep

router = APIRouter(prefix="/collections")


@router.put("/")
async def add_collection(
    conn: PgConnectionDep, collection: CollectionCreate
) -> Collection:
    """Create a collection."""
    dimensions = await get_dimensions(conn, collection.text_embedding_model)
    logger.info("Dimensions: {}", dimensions)
    async with conn.transaction():
        result = await conn.fetchrow(
            """
            INSERT INTO collection (
                name,
                text_embedding_model,
                text_distance_metric
            ) VALUES ($1, $2, $3)
            RETURNING id, name, text_embedding_model, text_distance_metric
            """,
            collection.name,
            collection.text_embedding_model,
            collection.text_distance_metric.value,
        )

        # Create a separate *partial* index on each collection.
        # This allows us to define different dimensions (and vector distance) for
        # each collection.
        #
        # https://github.com/pgvector/pgvector?tab=readme-ov-file#can-i-store-vectors-with-different-dimensions-in-the-same-column
        id = result["id"]
        vector_ops = collection.text_distance_metric.vector_ops()
        await conn.execute(
            f"""
            CREATE INDEX embedding_collection_{id}_index
            ON embedding
            USING hnsw ((embedding::vector({dimensions})) {vector_ops})
            WHERE collection_id = {id}
            """
        )
    return Collection.model_validate(dict(result))


@router.get("/")
async def list_collections(conn: PgConnectionDep) -> List[Collection]:
    """List collections."""
    results = await conn.fetch("SELECT id, name, text_embedding_model FROM collection")
    return [Collection.model_validate(dict(result)) for result in results]


PathCollectionId = Annotated[int, Path(..., description="The collection ID.")]


@router.get("/{id}")
async def get_collection(id: PathCollectionId, conn: PgConnectionDep) -> Collection:
    """Get a specific collection."""
    result = await conn.fetchrow(
        """
        SELECT id, name, text_embedding_model
        FROM collection
        WHERE id = $1
        """,
        id,
    )
    return Collection.model_validate(dict(result))



from fastapi import APIRouter

from dewy.chunk.router import router as chunks_router
from dewy.collection.router import router as collections_router
from dewy.document.router import router as documents_router

api_router = APIRouter(prefix="/api")

api_router.include_router(collections_router)
api_router.include_router(documents_router)
api_router.include_router(chunks_router)



from enum import Enum

from pydantic import BaseModel, ConfigDict, Field, TypeAdapter


class DistanceMetric(Enum):
    cosine = "cosine"
    inner_product = "ip"
    l2 = "l2"

    def vector_ops(self) -> str:
        match self:
            case DistanceMetric.cosine:
                return "vector_cosine_ops"
            case DistanceMetric.inner_product:
                return "vector_ip_ops"
            case DistanceMetric.l2:
                return "vector_l2_ops"

    def order_by(self, haystack: str, needle: str) -> str:
        match self:
            case DistanceMetric.cosine:
                return f"{haystack} <=> {needle}"
            case DistanceMetric.inner_product:
                return f"{haystack} <#> {needle}"
            case DistanceMetric.l2:
                return f"{haystack} <-> {needle}"

    def distance(self, haystack: str, needle: str) -> str:
        match self:
            case DistanceMetric.cosine:
                return f"1 - ({haystack} <=> {needle})"
            case DistanceMetric.inner_product:
                return f"({haystack} <#> {needle}) * -1"
            case DistanceMetric.l2:
                return f"{haystack} <-> {needle}"


class Collection(BaseModel):
    model_config = ConfigDict(from_attributes=True)

    """A collection of indexed documents."""
    id: int
    """The ID of the collection."""

    name: str
    """The name of the collection."""

    text_embedding_model: str
    """The name of the embedding model.

    NOTE: Changing embedding models is not currently supported.
    """

    text_distance_metric: DistanceMetric = DistanceMetric.cosine
    """The distance metric to use on the text embedding.

    NOTE: Changing distance metrics is not currently supported."""


collection_validator = TypeAdapter(Collection)


class CollectionCreate(BaseModel):
    """The request to create a collection."""

    name: str = Field(examples=["my_collection"])
    """The name of the collection."""

    text_embedding_model: str = Field(
        "openai:text-embedding-ada-002",
        examples=["openai:text-embedding-ada-002", "hf:BAAI/bge-small-en"],
    )
    """The name of the embedding model.

    NOTE: Changing embedding models is not currently supported.
    """

    text_distance_metric: DistanceMetric = DistanceMetric.cosine
    """The distance metric to use on the text embedding.

    NOTE: Changing distance metrics is not currently supported."""



from enum import Enum


class Environment(str, Enum):
    LOCAL = "LOCAL"
    STAGING = "STAGING"
    TESTING = "TESTING"
    PRODUCTION = "PRODUCTION"

    @property
    def is_debug(self) -> bool:
        return self in (self.LOCAL, self.STAGING, self.TESTING)

    @property
    def is_testing(self) -> bool:
        return self == self.TESTING

    @property
    def is_deployed(self) -> bool:
        return self in (self.STAGING, self.PRODUCTION)



from typing import Annotated, List

import asyncpg
from fastapi import APIRouter, BackgroundTasks, Path, Query

from dewy.common.collection_embeddings import CollectionEmbeddings
from dewy.common.db import PgConnectionDep, PgPoolDep
from dewy.document.models import Document

from .models import AddDocumentRequest

router = APIRouter(prefix="/documents")


async def ingest_document(document_id: int, pg_pool: asyncpg.Pool) -> None:
    url, embeddings = await CollectionEmbeddings.for_document_id(pg_pool, document_id)
    await embeddings.ingest(document_id, url)


@router.put("/")
async def add_document(
    pg_pool: PgPoolDep,
    background: BackgroundTasks,
    req: AddDocumentRequest,
) -> Document:
    """Add a document."""

    async with pg_pool.acquire() as conn:
        row = None
        row = await conn.fetchrow(
            """
        INSERT INTO document (collection_id, url, ingest_state)
        VALUES ($1, $2, 'pending')
        RETURNING id, collection_id, url, ingest_state, ingest_error
        """,
            req.collection_id,
            req.url,
        )

    document = Document.model_validate(dict(row))
    background.add_task(ingest_document, document.id, pg_pool)
    return document


PathDocumentId = Annotated[int, Path(..., description="The document ID.")]


@router.get("/")
async def list_documents(
    conn: PgConnectionDep,
    collection_id: Annotated[
        int | None,
        Query(description="Limit to documents associated with this collection"),
    ] = None,
) -> List[Document]:
    """List documents."""
    # TODO: Test
    if collection_id is None:
        results = await conn.fetch(
            """
            SELECT id, collection_id, url, ingest_state, ingest_error
            FROM document
        """
        )
    else:
        results = await conn.fetch(
            """
            SELECT id, collection_id, url, ingest_state, ingest_error
            FROM document WHERE collection_id = $1
        """,
            collection_id,
        )
    return [Document.model_validate(dict(result)) for result in results]


@router.get("/{id}")
async def get_document(conn: PgConnectionDep, id: PathDocumentId) -> Document:
    # TODO: Test / return not found?
    result = await conn.fetchrow(
        """
        SELECT id, collection_id, url, ingest_state, ingest_error
        FROM document WHERE id = $1
        """,
        id,
    )
    return Document.model_validate(dict(result))



from dataclasses import dataclass

from fastapi import HTTPException, status
from loguru import logger


@dataclass
class ExtractResult:
    text: str
    """The extracted text."""

    def is_empty(self) -> bool:
        if self.text:
            return False
        else:
            return True


def extract_from_pdf(
    local_path: str, *, extract_tables: bool = False, extract_images: bool = False
) -> ExtractResult:
    """Extract documents from a PDF."""

    logger.debug("Extracting from PDF '{}'", local_path)

    texts = []
    tables = []
    import fitz

    doc = fitz.open(local_path)
    for page in doc.pages():
        texts.append(page.get_text(sort=True))

        # TODO: Make the table analysis work using PyMuPDF or look at using
        # Table Transformer
        # (https://docs.llamaindex.ai/en/stable/examples/multi_modal/multi_modal_pdf_tables.html)
        if extract_tables:
            for table in page.find_tables():
                # TODO: join tables spanning multiple pages?
                # https://github.com/pymupdf/PyMuPDF-Utilities/blob/master/table-analysis/join_tables.ipynb
                df = table.to_pandas()
                table = df.to_csv(index=False)
                tables.extend(table)

        # TODO: Make image extraction work.
        if extract_images:
            for image in page.get_image_info(hashes=True, xrefs=True):
                print(f"Image: {image}")

    # TODO: Create image nodes for the tables and text.
    # TODO: Test that the document store is populated.
    # TODO: Test that the URL is included in the metadata.

    text = "".join(texts)
    return ExtractResult(text=text)


async def extract(
    url: str, *, extract_tables: bool = False, extract_images: bool = False
) -> ExtractResult:
    """Extract documents from a local or remote URL."""
    import httpx

    async with httpx.AsyncClient(follow_redirects=True) as client:
        # Determine the extension by requesting the headers.
        response = await client.head(url)
        response.raise_for_status()
        content_type = response.headers["content-type"]
        logger.debug("Content type of {} is {}", url, content_type)

        # Load the content.
        if content_type.startswith("application/pdf"):
            from tempfile import NamedTemporaryFile

            with NamedTemporaryFile(suffix=".pdf") as temp_file:
                logger.debug("Downloading {} to {}", url, temp_file.name)
                response = await client.get(url)
                response.raise_for_status()
                temp_file.write(response.content)

                return extract_from_pdf(
                    temp_file.name,
                    extract_tables=extract_tables,
                    extract_images=extract_images,
                )
        else:
            raise HTTPException(
                status_code=status.HTTP_415_UNSUPPORTED_MEDIA_TYPE,
                detail=f"Cannot add document from content-type '{content_type}'",
            )



from typing import List, Self, Tuple

import asyncpg
from llama_index.embeddings import BaseEmbedding
from llama_index.node_parser import SentenceSplitter
from llama_index.schema import TextNode
from loguru import logger

from dewy.chunk.models import TextResult
from dewy.collection.models import DistanceMetric
from dewy.config import settings

from .extract import extract


class CollectionEmbeddings:
    """Helper class for working with the embeddings in a collection."""

    def __init__(
        self,
        pg_pool: asyncpg.Pool,
        *,
        collection_id: int,
        text_embedding_model: str,
        text_embedding_dimensions: int,
        text_distance_metric: DistanceMetric,
    ) -> None:
        """Create a new CollectionEmbeddings."""
        self._pg_pool = pg_pool
        self.collection_id = collection_id
        self.text_embedding_model = text_embedding_model
        self.text_distance_metric = text_distance_metric

        self.extract_tables = False
        self.extract_images = False

        # TODO: Look at a sentence window splitter?
        self._splitter = SentenceSplitter()
        self._embedding = _resolve_embedding_model(self.text_embedding_model)

        field = f"embedding::vector({text_embedding_dimensions})"

        # TODO: Figure out how to limit by the number of *chunks* not the number
        # of embeddings.
        self._retrieve_embeddings = f"""
        SELECT
          chunk_id,
          {self.text_distance_metric.distance(field, "$2")} AS score
        FROM embedding
        WHERE collection_id = $1
        ORDER BY {self.text_distance_metric.order_by(field, "$2")}
        LIMIT $3
        """

        self._retrieve_chunks = f"""
        WITH relevant_embeddings AS (
          SELECT
            chunk_id,
            {self.text_distance_metric.distance(field, "$2")} AS score
          FROM embedding
          WHERE collection_id = $1
          ORDER BY {self.text_distance_metric.order_by(field, "$2")}
        )
        SELECT
          relevant_embeddings.chunk_id AS chunk_id,
          chunk.text AS text,
          relevant_embeddings.score AS score,
          chunk.document_id AS document_id
        FROM relevant_embeddings
        JOIN chunk
        ON chunk.id = relevant_embeddings.chunk_id
        LIMIT $3
        """

    @staticmethod
    async def for_collection_id(pg_pool: asyncpg.Pool, collection_id: int) -> Self:
        """Retrieve the collection embeddings of the given collection."""
        async with pg_pool.acquire() as conn:
            result = await conn.fetchrow(
                """
                SELECT
                    collection.id as id,
                    text_embedding_model,
                    text_distance_metric,
                    text_embedding_dimensions.dimensions AS text_embedding_dimensions
                FROM collection
                JOIN text_embedding_dimensions
                    ON text_embedding_dimensions.name = collection.text_embedding_model
                WHERE collection.id = $1;
                """,
                collection_id,
            )

            return CollectionEmbeddings(
                pg_pool,
                collection_id=result["id"],
                text_embedding_model=result["text_embedding_model"],
                text_embedding_dimensions=result["text_embedding_dimensions"],
                text_distance_metric=DistanceMetric(result["text_distance_metric"]),
            )

    @staticmethod
    async def for_document_id(pg_pool: asyncpg.Pool, document_id: int) -> (str, Self):
        """Retrieve the collection embeddings and the URL of the given document."""

        # TODO: Ideally the collection embeddings would be cached, and this
        # wouldn't need to exist.
        async with pg_pool.acquire() as conn:
            result = await conn.fetchrow(
                """
                SELECT
                    document.url as url,
                    collection.name,
                    collection.id as id,
                    collection.text_embedding_model,
                    collection.text_distance_metric,
                    text_embedding_dimensions.dimensions AS text_embedding_dimensions
                FROM document
                JOIN collection ON document.collection_id = collection.id
                JOIN text_embedding_dimensions
                    ON text_embedding_dimensions.name = collection.text_embedding_model
                WHERE document.id = $1;
                """,
                document_id,
            )

            # TODO: Cache the configured ingestions, and only recreate when needed?
            configured_ingestion = CollectionEmbeddings(
                pg_pool,
                collection_id=result["id"],
                text_embedding_model=result["text_embedding_model"],
                text_embedding_dimensions=result["text_embedding_dimensions"],
                text_distance_metric=DistanceMetric(result["text_distance_metric"]),
            )
            return (result["url"], configured_ingestion)

    async def retrieve_text_embeddings(
        self, query: str, n: int = 10
    ) -> List[Tuple[int, float]]:
        """Retrieve embeddings related to the given query.

        Parameters:
        - query: The query to retrieve matching embeddings for.
        - n: The number of embeddings to retrieve.

        Returns:
        List of `(chunk_id, score)` pairs from the embeddings.
        """
        embedded_query = await self._embedding.aget_text_embedding(query)

        async with self._pg_pool.acquire() as conn:
            logger.info("Executing SQL query for chunks from {}", self.collection_id)
            embeddings = await conn.fetch(
                self._retrieve_embeddings, self.collection_id, embedded_query, n
            )
            embeddings = [e["chunk_id"] for e in embeddings]
            return embeddings

    async def retrieve_text_chunks(self, query: str, n: int = 10) -> List[TextResult]:
        """Retrieve embeddings related to the given query.

        Parameters:
        - query: The query to retrieve matching embeddings for.
        - n: The number of embeddings to retrieve.

        Returns:
        List of chunk_ids from the embeddings.
        """
        embedded_query = await self._embedding.aget_text_embedding(query)

        async with self._pg_pool.acquire() as conn:
            logger.info("Executing SQL query for chunks from {}", self.collection_id)
            embeddings = await conn.fetch(
                self._retrieve_chunks, self.collection_id, embedded_query, n
            )
            embeddings = [
                TextResult(
                    chunk_id=e["chunk_id"],
                    document_id=e["document_id"],
                    score=e["score"],
                    text=e["text"],
                    raw=True,
                    start_char_idx=None,
                    end_char_idx=None,
                )
                for e in embeddings
            ]
            return embeddings

    async def ingest(self, document_id: int, url: str) -> None:
        logger.info("Loading content for document {} from '{}'", document_id, url)
        extracted = await extract(
            url, extract_tables=self.extract_tables, extract_images=self.extract_images
        )
        if extracted.is_empty():
            logger.error(
                "No content retrieved from for document {} from '{}'", document_id, url
            )
            return

        logger.info(
            "Chunking text of length {} for {}", len(extracted.text), document_id
        )

        # Extract chunks (snippets) and perform the direct embedding.
        text_chunks = await self._chunk_sentences(extracted.text)

        logger.info("Chunking produced {} chunks for {}", len(text_chunks), document_id)

        # TODO: support non-text chunks
        # TODO: support non-snippet text chunks (eg., summary values)
        # TODO: support indirect embeddings
        async with self._pg_pool.acquire() as conn:
            async with conn.transaction():

                def encode_chunk(c: str) -> str:
                    # We believe that either invalid unicode or the occurrence
                    # of nulls was causing problems that *looked* like only the
                    # first page from a PDF was being indexed
                    # (https://github.com/DewyKB/dewy/issues/20). We do not know
                    # that all of this is truly necessary.
                    encoded = c.encode("utf-8").decode("utf-8", "ignore")
                    return encoded.replace("\x00", "\uFFFD")

                # First, insert the chunks.
                await conn.executemany(
                    """
                    INSERT INTO chunk (document_id, kind, text)
                    VALUES ($1, $2, $3);
                    """,
                    [
                        (document_id, "text", encode_chunk(text_chunk))
                        for text_chunk in text_chunks
                    ],
                )

                # Then, embed each of those chunks.
                # We assume no chunks for the document existed before, so we can iterate
                # over the chunks.
                chunks = conn.cursor(
                    "SELECT id, text FROM chunk WHERE document_id = $1", document_id
                )

                # TODO: Write this loop in a cleaner async way, to limit the number of
                # in-flight requests as well as batching up the embedding requests.
                # Currently, this uses Llama Index embeddings, which requires we put
                # all the texts to embed in a list.
                #
                # Ideally, we could take a chunk of embeddings, embed them, and then
                # start writing that to the DB asynchronously.
                embedding_chunks = [
                    (chunk["id"], chunk["text"]) async for chunk in chunks
                ]

                # Extract just the text and embed it.
                logger.info(
                    "Computing {} embeddings for {}", len(embedding_chunks), document_id
                )
                embeddings = await self._embedding.aget_text_embedding_batch(
                    [item[1] for item in embedding_chunks]
                )

                # Change the shape to a list of triples (for writing to the DB)
                embeddings = [
                    (self.collection_id, chunk_id, chunk_text, embedding)
                    for (chunk_id, chunk_text), embedding in zip(
                        embedding_chunks, embeddings
                    )
                ]

                logger.info(
                    "Writing {} embeddings for {}", len(embeddings), document_id
                )
                await conn.executemany(
                    """
                    INSERT INTO embedding (collection_id, chunk_id, key_text, embedding)
                    VALUES ($1, $2, $3, $4)
                    """,
                    embeddings,
                )
                logger.info("Wrote {} embeddings for {}", len(embeddings), document_id)

                await conn.execute(
                    """
                    UPDATE document
                    SET ingest_state = 'ingested', ingest_error = NULL
                    WHERE id = $1
                    """,
                    document_id,
                )

    async def _chunk_sentences(self, text: str) -> List[str]:
        # This uses llama index a bit oddly. Unfortunately:
        #  - It returns `BaseNode` even though we know these are `TextNode`
        #  - It returns a `List` rather than an `Iterator` / `Generator`, so
        #    all resulting nodes are resident in memory.
        #  - It uses metadata to return the "window" (if using sentence windows).
        return [node.text for node in await self._splitter.acall([TextNode(text=text)])]


DEFAULT_OPENAI_EMBEDDING_MODEL: str = "openai:text-embedding-ada-002"
DEFAULT_HF_EMBEDDING_MODEL: str = "hf:BAAI/bge-small-en"


async def get_dimensions(conn: asyncpg.Connection, model_name: str) -> int:
    dimensions = await conn.fetchval(
        """
        SELECT dimensions
        FROM text_embedding_dimensions
        WHERE name = $1
        """,
        model_name,
    )

    if dimensions is not None:
        return dimensions

    model = _resolve_embedding_model(model_name)
    dimensions = len(await model.aget_text_embedding("test string"))

    # TODO: Deal with concurrency? I suspect it is OK if this fails
    # due to the uniqueness constraint, and we should just move on.
    # Someone wrote the value for that name to the table, and we should
    # have determined the same values.
    await conn.execute(
        """
        INSERT INTO text_embedding_dimensions (name, dimensions)
        VALUES ($1, $2)
        """,
        model_name,
        dimensions,
    )

    return dimensions


def _resolve_embedding_model(model: str) -> BaseEmbedding:
    if not model:
        if settings.OPENAI_API_KEY:
            model = DEFAULT_OPENAI_EMBEDDING_MODEL
        else:
            model = DEFAULT_HF_EMBEDDING_MODEL

    split = model.split(":", 2)
    if split[0] == "openai":
        from llama_index.embeddings import OpenAIEmbedding

        return OpenAIEmbedding(model=split[1])
    elif split[0] == "hf":
        from llama_index.embeddings import HuggingFaceEmbedding

        return HuggingFaceEmbedding(model_name=split[1])
    else:
        raise ValueError(f"Unrecognized embedding model '{model}'")



import hashlib
import os
import re
from typing import Dict, Iterable, Optional, Tuple

import asyncpg
from loguru import logger


async def apply_migrations(
    conn: asyncpg.Connection, *, migration_dir: str = "migrations"
):
    """Applies the migrations to the database."""

    # TODO: For deployments, we may wish to make this a script that
    # can be executed, rather than always calling it on startup.

    # Ensure the migration table exists.
    await conn.execute(
        """
        CREATE TABLE IF NOT EXISTS migration (
            -- Record the versions we have applied.
            id INTEGER NOT NULL,
            -- Record the SHA of the version so we can detect changes.
            sha256 TEXT,
            PRIMARY KEY (id)
        );
        """
    )

    defined_migrations = _get_defined_migrations(migration_dir)
    applied_migrations = await _get_applied_migrations(conn)

    applied = 0
    for migration_id, migration_file in defined_migrations:
        applied_sha256 = applied_migrations.pop(migration_id, None)
        if await _apply_migration(conn, migration_id, migration_file, applied_sha256):
            applied += 1

    if applied_migrations:
        logger.warn("Unrecognized migrations applied: {}", applied_migrations)

    logger.info(
        "Migrations complete. {} total, {} newly applied",
        len(defined_migrations),
        applied,
    )


MIGRATION_RE = re.compile(r"([0-9]{4})[a-zA-Z0-9_-]+\.sql")


def _get_defined_migrations(dir: str) -> Iterable[Tuple[int, str]]:
    files = {}
    for file in os.listdir(dir):
        match = MIGRATION_RE.fullmatch(file)
        if match:
            number = int(match.group(1))
            existing_file = files.setdefault(number, file)
            if existing_file != file:
                raise ValueError(
                    f"Multiple migrations for {number}: '{existing_file}' and '{file}'"
                )
        else:
            logger.warning(
                "File '{}' in migrations did not match migration naming '{}'",
                file,
                MIGRATION_RE.pattern,
            )
    return [(id, os.path.join(dir, file)) for (id, file) in sorted(files.items())]


async def _get_applied_migrations(conn: asyncpg.Connection) -> Dict[int, str]:
    applied = await conn.fetch("SELECT id, sha256 FROM migration ORDER BY id")
    return {row["id"]: row["sha256"] for row in applied}


async def _apply_migration(
    conn: asyncpg.Connection,
    migration_id: int,
    migration_path: str,
    applied_sha256: Optional[str],
) -> bool:
    """Apply the given migration if it hasn't already been applied.

    Uses `applied_sha256` to determine whether it has been applied.
    """
    # NOTE: Using the sha256 from the list means that other processes
    # may have applied the migration. If we expect to run the migrations
    # concurrently, we should have a more explicit locking. On the other
    # hand, in a "production" environment, we probably want to explicitly
    # apply the migrations, rather than having every node try to apply
    # them on startup -- so we could just run them as part of the deployment
    # script, which likely means we *don't* actually run these concurrently.

    with open(migration_path) as migration_file:
        migration_content = migration_file.read()

        pending_sha256 = hashlib.sha256(migration_content.encode("utf8")).hexdigest()

        async with conn.transaction():
            if pending_sha256 == applied_sha256:
                logger.debug(
                    "Migration '{}' already applied with sha {}",
                    migration_path,
                    pending_sha256,
                )
                return False
            elif applied_sha256 is not None:
                raise ValueError(
                    f"'{migration_path}' applied with different SHA. Recreate DB."
                )
            else:
                logger.info("Applying migration '{}'", migration_path)

            # Apply the migration.
            await conn.execute(migration_content)
            await conn.execute(
                "INSERT INTO migration (id, sha256) VALUES ($1, $2);",
                migration_id,
                pending_sha256,
            )

            return True



from typing import Annotated, List

from fastapi import APIRouter, Path, Query

from dewy.common.collection_embeddings import CollectionEmbeddings
from dewy.common.db import PgPoolDep

from .models import Chunk, RetrieveRequest, RetrieveResponse, TextChunk

router = APIRouter(prefix="/chunks")


@router.get("/")
async def list_chunks(
    pg_pool: PgPoolDep,
    collection_id: Annotated[
        int | None, Query(description="Limit to chunks associated with this collection")
    ] = None,
    document_id: Annotated[
        int | None, Query(description="Limit to chunks associated with this document")
    ] = None,
    page: int | None = 1,
    perPage: int | None = 10,
) -> List[Chunk]:
    """List chunks."""

    # TODO: handle collection & document ID
    results = await pg_pool.fetch(
        """
        SELECT chunk.id, chunk.document_id, chunk.kind, TRUE as raw, chunk.text
        FROM chunk
        JOIN document ON document.id = chunk.document_id
        WHERE document.collection_id = coalesce($1, document.collection_id)
        AND chunk.document_id = coalesce($2, chunk.document_id)
        ORDER BY chunk.id
        OFFSET $4
        LIMIT $3
        """,
        collection_id,
        document_id,
        perPage,
        page,
    )
    return [TextChunk.model_validate(dict(result)) for result in results]


PathChunkId = Annotated[int, Path(..., description="The chunk ID.")]


@router.get("/{id}")
async def get_chunk(
    pg_pool: PgPoolDep,
    id: PathChunkId,
) -> Chunk:
    # TODO: Test / return not found?
    result = await pg_pool.fetchrow(
        """
        SELECT id, document_id, kind, text
        FROM chunk WHERE id = $1
        """,
        id,
    )
    return Chunk.model_validate(dict(result))


@router.post("/retrieve")
async def retrieve_chunks(
    pg_pool: PgPoolDep, request: RetrieveRequest
) -> RetrieveResponse:
    """Retrieve chunks based on a given query."""

    # TODO: Revisit response synthesis and hierarchical fetching.

    collection = await CollectionEmbeddings.for_collection_id(
        pg_pool, request.collection_id
    )
    text_results = await collection.retrieve_text_chunks(
        query=request.query, n=request.n
    )

    return RetrieveResponse(
        summary=None,
        text_results=text_results if request.include_text_chunks else [],
        image_results=[],
    )



import contextlib
from typing import AsyncIterator, TypedDict
from pathlib import Path
import os

import uvicorn
import asyncpg
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fastapi.staticfiles import StaticFiles
from loguru import logger

from dewy.common import db
from dewy.common.db_migration import apply_migrations
from dewy.config import app_configs, settings
from dewy.routes import api_router


class State(TypedDict):
    pg_pool: asyncpg.Pool

# Resolve paths, independent of PWD
current_file_path = Path(__file__).resolve()
react_build_path = current_file_path.parent.parent / 'frontend' / 'dist'
migrations_path = current_file_path.parent.parent / 'migrations'

@contextlib.asynccontextmanager
async def lifespan(_app: FastAPI) -> AsyncIterator[State]:
    """Function creating instances used during the lifespan of the service."""

    # TODO: Look at https://gist.github.com/mattbillenstein/270a4d44cbdcb181ac2ed58526ae137d
    # for simple migration scripts.
    async with db.create_pool(settings.DB.unicode_string()) as pg_pool:
        if settings.APPLY_MIGRATIONS:
            async with pg_pool.acquire() as conn:
                await apply_migrations(conn, migration_dir=migrations_path)

        logger.info("Created database connection")
        state = {
            "pg_pool": pg_pool,
        }
        yield state


app = FastAPI(lifespan=lifespan, **app_configs)

origins = [
    "http://localhost:5173",
]
app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


@app.get("/healthcheck", include_in_schema=False)
async def healthcheck() -> dict[str, str]:
    return {"status": "ok"}


app.include_router(api_router)

if settings.SERVE_ADMIN_UI and os.path.isdir(react_build_path):
    logger.info("Running admin UI at http://localhost/admin")
    # Serve static files from the React app build directory
    app.mount("/admin", StaticFiles(directory=str(react_build_path), html=True), name="static")

# Function for running Dewy as a script
def run(*args):
   uvicorn.run("dewy.main:app", host="0.0.0.0", port=80)


from typing import Annotated, Literal, Optional, Sequence, Union

from pydantic import BaseModel, Field


class TextChunk(BaseModel):
    id: int
    document_id: int
    kind: Literal["text"] = "text"
    text: str

    raw: bool
    text: str
    start_char_idx: Optional[int] = Field(
        default=None, description="Start char index of the chunk."
    )
    end_char_idx: Optional[int] = Field(
        default=None, description="End char index of the chunk."
    )


class ImageChunk(BaseModel):
    id: int
    document_id: int
    kind: Literal["image"] = "image"

    image: Optional[str] = Field(..., description="Image of the node.")
    image_mimetype: Optional[str] = Field(..., description="Mimetype of the image.")
    image_path: Optional[str] = Field(..., description="Path of the image.")
    image_url: Optional[str] = Field(..., description="URL of the image.")


Chunk = Annotated[Union[TextChunk, ImageChunk], Field(discriminator="kind")]


class RetrieveRequest(BaseModel):
    """A request for retrieving chunks from a collection."""

    collection_id: int
    """The collection to retrieve chunks from."""

    query: str
    """The query string to use for retrieval."""

    n: int = 10
    """The number of chunks to retrieve."""

    # TODO: We may want an option which only excludes nodes included in the summary.
    # For instance -- if we summarize the text statements, maybe it only includes
    # images and tables in the response. But for now, this is a big switch to
    # exclude statements entirely.
    include_text_chunks: bool = True
    """Whether to include text chunks in the result.

    If this is false, no text chunks will be included in the result, although
    the summary (if enbaled) may include information from the chunks.
    """

    include_image_chunks: bool = True
    """Whether to include image chunks in the result.

    If this is false, no image chunks will be included in the result, although
    the summary (if enbaled) may include information from the chunks.
    """

    include_summary: bool = False
    """Whether to include a generated summary."""


class TextResult(BaseModel):
    chunk_id: int
    """The ID of the chunk associated with this result"""

    document_id: int
    """The ID of the document associated with this result"""

    score: float
    """The similarity score of this result."""

    text: str
    "Textual description of the chunk."

    raw: bool
    start_char_idx: Optional[int] = Field(
        default=None, description="Start char index of the chunk."
    )
    end_char_idx: Optional[int] = Field(
        default=None, description="End char index of the chunk."
    )


class ImageResult(BaseModel):
    chunk_id: int
    """The ID of the chunk associated with this result"""

    document_id: int
    """The ID of the document associated with this result"""

    score: float
    """The similarity score of this result."""

    image: Optional[str] = Field(..., description="Image of the node.")
    image_mimetype: Optional[str] = Field(..., description="Mimetype of the image.")
    image_path: Optional[str] = Field(..., description="Path of the image.")
    image_url: Optional[str] = Field(..., description="URL of the image.")


class RetrieveResponse(BaseModel):
    """The response from a retrieval request."""

    summary: Optional[str]
    """Summary of the retrieved chunks."""

    text_results: Sequence[TextResult]
    """Retrieved text chunks."""

    image_results: Sequence[ImageResult]
    """Retrieved image chunks."""



from enum import Enum
from typing import Optional

from pydantic import BaseModel


class AddDocumentRequest(BaseModel):
    collection_id: Optional[int] = None
    """The id of the collection the document should be added to."""

    url: str
    """The URL of the document to add."""


class IngestState(Enum):
    PENDING = "pending"
    """Document is pending ingestion."""

    INGESTED = "ingested"
    """Document has been ingested."""

    FAILED = "failed"
    """Document failed to be ingested. See `ingest_errors` for details."""


class Document(BaseModel):
    """Schema for documents in the SQL DB."""

    id: Optional[int] = None
    collection_id: int

    url: str

    ingest_state: Optional[IngestState] = None
    ingest_error: Optional[str] = None
